@@ -1,4 +1,5 @@
 from collections import defaultdict
+from io import StringIO
 from xml.sax.saxutils import escape
 
 from hashedindex import HashedIndex
@@ -6,7 +7,6 @@ from hashedindex.textparser import NullStemmer, word_tokenize
 
 from hashedixsearch.analysis import SynonymAnalyzer
 from hashedixsearch._utils import (
-    _candidate_matches,
     _is_separator,
     _render_match,
 )
@@ -111,7 +111,7 @@ class HashedIXSearch(object):
             if frequency == doc_length:
                 return doc_id
 
-    def _ngram_to_term(self, ngram, case_sensitive):
+    def _next_token(self, ngram, case_sensitive):
         return next(
             self.tokenize(
                 doc=str().join(ngram),
@@ -119,15 +119,14 @@ class HashedIXSearch(object):
                 retain_punctuation=True,
                 retain_whitespace=True,
             )
-        )
+        )[0]
 
     def highlight(self, doc, terms, case_sensitive=True, term_attributes=None):
         # If no terms are provided to match on, do not attempt highlighting
         if not terms:
             return escape(doc)
 
-        terms = {term: list(term) for term in terms}
-        max_n = max(len(term) for term in terms.values())
+        max_n = max(len(term) for term in terms)
         term_attributes = term_attributes or {}
 
         # Generate unstemmed ngrams of the maximum term length
@@ -153,49 +152,44 @@ class HashedIXSearch(object):
         for n in range(1, max_n + 1):
             ngrams.append(final_ngram[n:])
 
+        # Remove the end-of-stream ngram
+        ngrams.pop()
+
         # Build up a marked-up representation of the original document
-        tag = None
-        markup = ""
-        accumulator = ""
+        candidates = {}
+        accumulator = StringIO()
+        markup = StringIO()
 
-        ngrams = iter(ngrams)
-        while ngram := next(ngrams):
-
-            # Stop when we reach an empty end-of-stream ngram
-            if not ngram:
-                break
-
-            # Determine whether any of the highlighting terms match
-            ngram_term = self._ngram_to_term(ngram, case_sensitive)
-
-            if not tag and not _is_separator(ngram_term[0]):
-                tag = _candidate_matches(ngram_term, terms)
-                markup += accumulator
-                accumulator = ""
+        for ngram in ngrams:
 
             # Consume one token at a time
-            accumulator += escape(ngram[0])
-            emit = tag is None
+            token = self._next_token(ngram, case_sensitive)
 
-            # Advance the match window of each candidate tag element
-            if tag and not _is_separator(ngram_term[0]):
-                tag = {
-                    term: tokens[1:]
-                    for term, tokens in tag.items()
-                    if tokens[0] == ngram_term[0]
+            # Advance the match window for each candidate term
+            if not _is_separator(token):
+                candidates = {
+                    term: term[1:]
+                    for term in candidates.values() or terms
+                    if term[0] == token
                 }
 
-                # Close the markup when any term's tokens have been consumed
-                closing_term = next(filter(lambda k: not tag[k], tag), None)
-                if closing_term:
-                    attributes = term_attributes.get(closing_term)
-                    accumulator = _render_match(accumulator, attributes)
-                    emit = True
-                    tag = None
+            # Prepare the current token for output, and write it to the
+            # accumulator buffer if we are within candidate highlight tokens
+            output = escape(ngram[0])
+            if candidates:
+                accumulator.write(output)
 
-            # Output accumulated tokens when we encounter emit-points
+            # Render highlight markup once a candidate's terms are consumed
+            emit = next(filter(lambda k: not candidates[k], candidates), None)
             if emit:
-                markup += accumulator
-                accumulator = ""
+                attributes = term_attributes.get(emit)
+                output = _render_match(accumulator, attributes)
+                candidates = {}
+                accumulator = StringIO()
 
-        return markup + accumulator
+            # Write output to the markup stream when match candidates are empty
+            if not candidates:
+                markup.write(output)
+
+        markup.write(accumulator.getvalue())
+        return markup.getvalue()
